/********************************************************************************
*
*	Files: animWindow.h and animWindow.cpp
*	Written by Steve Dutcher and Andy Gardner
*	
*	This class is the glWindow which draws the animation.  Quaternion camera
*	motion has been implemented in here.  Another thing to note is that the 
*	one important part of this class is that the user can get then data it needs
*	from this class in the whoIsShown array and then pass that into the constructor
*	for creating a clone of the animation window so that the user can see two
*	views at the same time
*
*	NOTE: read animWindow.h for a description of functions
*
**********************************************************************************/

#include <stdio.h>
#include <math.h>

#include "animWindow.h"
#include "3DUtils.h"

#ifndef PI
#define PI (3.14159265358979323)
#endif

enum MARKER_TYPE { ORIGINAL=0, EULER, QUATERNION };


animWindow::animWindow(int x, int y, int w, int h, motionData *md, int* shownValues) :
	Fl_Gl_Window(x,y,w,h)
{
	mdata = md;

	if( shownValues == NULL ){
		for(int i=0; i<9; i++)
			whoIsShown[i] = 0;
		whoIsShown[ EULER_MARKER ] = 1;
		whoIsShown[ EULER_LINE ] = 1;
	}else{
		for(int i=0; i<9; i++)
			whoIsShown[i] = shownValues[i];
	}

	// pick a starting place for the camera
	rotateCX = 20   ;
	rotateCY = -40  ;
	scaleC   = -300 ;
	fieldOfView = 45;

	lookAtX = lookAtY = lookAtZ = 0 ;

	frame = 1;

#ifndef _LINUX
	// This has been unincluded from the linux version becuase our linux machines didn't have
	// hardware support for openGL which prevented us from using this feature
	glDepthFunc(GL_LEQUAL)  ;
	glDepthMask(GL_TRUE)    ;
#endif
	((RunSlider *) time)->addWidgetToDamageList( this );
}

animWindow::~animWindow(){
	((RunSlider *) time)->removeWidgetFromDamageList( this );
}

void animWindow::attachToSlider(){
	((RunSlider *) time)->addWidgetToDamageList( this );
}
void animWindow::detachFromSlider(){
	((RunSlider *) time)->removeWidgetFromDamageList( this );
}

// This function gets called by the button callback and will set the value of an
// item in setShown to their value.  You should uses names from the enum in 
// animWindow called SHOWN_DATA_TYPE
void animWindow::setShown(int value, int dataType)
{
	whoIsShown[dataType] = value;
}

// Called when new data is opened - mainly for initializing and locking the time slider
// to the current frame.
void animWindow::openMotionDataFile( char* filename )
{
	if(filename == NULL || filename[0] == '\0')
		return;
	mdata->clearMotionData();
	mdata->readBVHfile( filename );

	((RunSlider *) time)->RunLength(mdata->getNumberOfFrames()) ;
	((RunSlider *) time)->value(1);
	((RunButton *) play)->setFrameTime( mdata->getFrameTime() );
	damage(1);
}

// Called when new data is opened - mainly for initializing and locking the time slider
// to the current frame.
void animWindow::openAcclaimDataFiles( char* filenameASF, char* filenameAMC )
{
	if(filenameASF == NULL || filenameASF[0] == '\0')
		return;
	if(filenameAMC == NULL || filenameAMC[0] == '\0')
		return;
	mdata->clearMotionData();
	mdata->readAcclaimFiles( filenameASF, filenameAMC );

	((RunSlider *) time)->RunLength(mdata->getNumberOfFrames()) ;
	printf ("%d\n", (mdata->getNumberOfFrames())) ;
	((RunSlider *) time)->value(1);
	damage(1);
}

// draw the windows proceeds as follows:
// first, clear the window (be sure to clear the Z-buffer)
// then set up the projection and view
// then draw the ground plane
// then set up for shadows and draw
// then draw the objects
void animWindow::draw()
{
	// clear the window, be sure to clear the Z-Buffer too
	glClearColor(0,0,0,0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glEnable(GL_DEPTH_TEST) ;

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  
	// set up the drawing transforms
	glMatrixMode(GL_PROJECTION);
	glViewport(0,0,w(),h());
	glLoadIdentity();
	// compute the aspect ratio so we don't distort things
	double aspect = ((double) w()) / ((double) h());
	gluPerspective(fieldOfView, aspect, .1, 10000);

	// put the camera where we want it to be
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// We are looking from the CAMERAVIEW so use the eye and look values
	// gluLookAt(eyeX, eyeY, eyeZ, lookX, lookY, lookZ, 0,1,0);
	glTranslatef(0.0f, 0.0f, scaleC) ;
	glRotatef(rotateCX, 1.0f, 0.0f, 0.0f) ;
	glRotatef(rotateCY, 0.0f, 1.0f, 0.0f) ;
	glTranslatef(lookAtX, lookAtY, lookAtZ) ;

	//Here is where we tell the rib function where the camera is...
	float rotationMatrix[16] ;
	glGetFloatv(GL_MODELVIEW_MATRIX, rotationMatrix);
	//printf("GL's matrix\n");
	//printf("%f\t%f\t%f\t%f\n%f\t%f\t%f\t%f\n%f\t%f\t%f\t%f\n%f\t%f\t%f\t%f\n", rotationMatrix[0],rotationMatrix[1],rotationMatrix[2],rotationMatrix[3],rotationMatrix[4],rotationMatrix[5],rotationMatrix[6],rotationMatrix[7],rotationMatrix[8],rotationMatrix[9],rotationMatrix[10],rotationMatrix[11],rotationMatrix[12],rotationMatrix[13],rotationMatrix[14],rotationMatrix[15]);
	
	mdata->setCameraPosition( rotationMatrix );

	glShadeModel(GL_SMOOTH) ;

	drawFloor( mdata->getFloorHeight() );

	drawMotionData( 0 );

	// dropShadowMatrix() multilplies a matrix on the stack which will cause everything
	// after it to be drawn on the floor as shadows
	dropShadowMatrix( mdata->getFloorHeight() );
	drawMotionData( 1 );
}

// This funciton draws the motion data the is selected int the variable whoIsShown
// It also passes an int shadow which is used for dropping the drop shadows so that
// they are the right color
void animWindow::drawMotionData( int shadow )
{
	if( mdata->doesExist() )
	{
		// Draw joint lines
		if( whoIsShown[ORIGINAL_LINE] == 1 )
			for(int l=0; l<mdata->getNumberOfRoots(); l++)
			{
				joint *j = mdata->getJoint( mdata->getRootName(l) );
				drawJointLine( j, ORIGINAL, shadow );
			}
		if( whoIsShown[EULER_LINE] == 1 )
			for(int l=0; l<mdata->getNumberOfRoots(); l++)
			{
				joint *j = mdata->getJoint( mdata->getRootName(l) );
				drawJointLine( j, EULER, shadow );
			}
		if( whoIsShown[QUATERNION_LINE] == 1 )
			for(int l=0; l<mdata->getNumberOfRoots(); l++)
			{
				joint *j = mdata->getJoint( mdata->getRootName(l) );
				drawJointLine( j, QUATERNION, shadow );
			}
	
		// Draw Ellipsoids
		if( whoIsShown[ORIGINAL_ELLIPSOID] == 1 )
			for(int l=0; l<mdata->getNumberOfRoots(); l++)
			{
				joint *j = mdata->getJoint( mdata->getRootName(l) );
				drawJointEllipsoid( j, ORIGINAL, shadow );
			}
		if( whoIsShown[EULER_ELLIPSOID] == 1 )
			for(int l=0; l<mdata->getNumberOfRoots(); l++)
			{
				joint *j = mdata->getJoint( mdata->getRootName(l) );
				drawJointEllipsoid( j, EULER, shadow );
			}
		if( whoIsShown[QUATERNION_ELLIPSOID] == 1 )
			for(int l=0; l<mdata->getNumberOfRoots(); l++)
			{
				joint *j = mdata->getJoint( mdata->getRootName(l) );
				drawJointEllipsoid( j, QUATERNION, shadow );
			}

		// Draw joint markers
		if (!shadow)
		{
			if( whoIsShown[ORIGINAL_MARKER] == 1 )
				drawMarkers( ORIGINAL, shadow );
			if( whoIsShown[EULER_MARKER] == 1 )
				drawMarkers( EULER, shadow );
			if( whoIsShown[QUATERNION_MARKER] == 1 )
				drawMarkers( QUATERNION, shadow );
		}
	}
}

// This function recursively draws lines to represent the joints
void animWindow::drawJointLine( joint *j, int markerType, int shadow )
{
	//If you're a shadow, make your color black, otherwise draw the proper lines.
	if( shadow )
		glColor4f( 0, 0, 0, 1 );
	else if( markerType == EULER )
		glColor4f( 1, 0, 0, 1 );
	else if( markerType == QUATERNION )
		glColor4f( 0, 1, 0, 1 );
	else if( markerType == ORIGINAL )
		glColor4f( 0, 0, 1, 1 );

	//End recursion if you're an effector.
	if( j->getType() == EFFECTOR ) 
		return;

	int sTime = (int)time->value();
	float mx, my, mz, mx2, my2, mz2;

	if( markerType == EULER )
		j->getEulerMarkerPosition( sTime ,mx, my, mz );
	else if( markerType == QUATERNION )
		j->getQuaternionMarkerPosition( sTime ,mx, my, mz );
	else if( markerType == ORIGINAL )
		j->getOriginalMarkerPosition( sTime ,mx, my, mz );

	//Recurse in a depth-first search order.
	for( int i=0; i<j->numOfChildren(); i++ )
	{
		joint *j2 = mdata->getJoint( j->getChild(i) );
		if( markerType == EULER )
			j2->getEulerMarkerPosition( sTime, mx2, my2, mz2 );
		else if( markerType == QUATERNION )
			j2->getQuaternionMarkerPosition( sTime, mx2, my2, mz2 );
		else if( markerType == ORIGINAL )
			j2->getOriginalMarkerPosition( sTime, mx2, my2, mz2 );
		
		drawLine( mx, my, mz, mx2, my2, mz2 );
	
		drawJointLine( j2, markerType, shadow );
	}
}

// This function recursively draws an ellipsoid to represent the joints
void animWindow::drawJointEllipsoid( joint *j, int markerType, int shadow )
{
	// Select the proper color to draw the ellipsoid.
	if( shadow )
		glColor4f( 0, 0, 0, 1 );
	else if( markerType == EULER )
		glColor4f( 1, 0, 1, 1 );
	else if( markerType == QUATERNION )
		glColor4f( 1, 1, 0, 1 );
	else if( markerType == ORIGINAL )
		glColor4f( 0, 1, 1, 1 );

	// End recursion if you're an end effector.
	if( j->getType() == EFFECTOR ) 
		return;

	int sTime = (int)time->value();
	float mx, my, mz, mx2, my2, mz2;

	if( markerType == EULER )
		j->getEulerMarkerPosition( sTime ,mx, my, mz );
	else if( markerType == QUATERNION )
		j->getQuaternionMarkerPosition( sTime ,mx, my, mz );
	else if( markerType == ORIGINAL )
		j->getOriginalMarkerPosition( sTime ,mx, my, mz );

	// Recurse with a depth-first recursion technique and draw -all- the joints.
	for( int i=0; i<j->numOfChildren(); i++ )
	{
		joint *j2 = mdata->getJoint( j->getChild(i) );
		if( markerType == EULER )
			j2->getEulerMarkerPosition( sTime, mx2, my2, mz2 );
		else if( markerType == QUATERNION )
			j2->getQuaternionMarkerPosition( sTime, mx2, my2, mz2 );
		else if( markerType == ORIGINAL )
			j2->getOriginalMarkerPosition( sTime, mx2, my2, mz2 );
		
		ellipsify( mx, my, mz, mx2, my2, mz2 );
	
		drawJointEllipsoid( j2, markerType, shadow );
	}
}

// This function draws boxes and trails at the markers for each end joint
void animWindow::drawMarkers(int markerType, int shadow)
{
	float mx, my, mz, mx2, my2, mz2;
	int sTime = (int)time->value();

	// Draw cubes at the markers
	mdata->resetJointPointer();
	for(int jcnt=0; jcnt<mdata->getNumberOfJoints(); jcnt++)
	{
		//Find the proper color...
		if( shadow )
			glColor4f( 0, 0, 0, 1 );
		else if( markerType == EULER )
			glColor4f( 0, 1, 0, 1 );
		else if( markerType == QUATERNION )
			glColor4f( 0, 0, 1, 1 );
		else if( markerType == ORIGINAL )
			glColor4f( 1, 0, 0, 1 );

		// If you're at the end of the skeleton (an effector), draw the markers and trails...
		joint *j = mdata->getCurrentJointPointer();
		if( j->getType() == EFFECTOR )
		{
			// This draws the markers
		if( markerType == EULER )
			j->getEulerMarkerPosition( sTime ,mx, my, mz );
		else if( markerType == QUATERNION )
			j->getQuaternionMarkerPosition( sTime ,mx, my, mz );
		else if( markerType == ORIGINAL )
			j->getOriginalMarkerPosition( sTime ,mx, my, mz );
			drawCube(mx, my, mz, 1);

			// This is a trace of the marker positions
			int i;
			if( sTime > 50 )
				i = sTime - 50;
			else
				i = 1;
			//This handles the fading of the lines.
			for(i; i<sTime; i++)
			{
				if( markerType == EULER )
					glColor4f( 0, 1, 0, 1.0f-((sTime-i)/50.0f) );
				else if( markerType == QUATERNION )
					glColor4f( 0, 0, 1, 1.0f-((sTime-i)/50.0f) );
				else if( markerType == ORIGINAL )
					glColor4f( 1, 0, 0, 1.0f-((sTime-i)/50.0f) );
				if( markerType == EULER ){
					j->getEulerMarkerPosition( i ,mx, my, mz );
					j->getEulerMarkerPosition( i+1, mx2, my2, mz2 );
				}else if( markerType == QUATERNION ){
					j->getQuaternionMarkerPosition( i ,mx, my, mz );
					j->getQuaternionMarkerPosition( i+1, mx2, my2, mz2 );
				}else if( markerType == ORIGINAL ){
					j->getOriginalMarkerPosition( i ,mx, my, mz );
					j->getOriginalMarkerPosition( i+1, mx2, my2, mz2 );
				}	
				drawLine(mx, my, mz, mx2, my2, mz2);
			}
		}
		// Simply step through the list for this function - don't recurse.
		mdata->advanceJointPointer();
	}	
}


// This code was given to be by Professor Michael Gleicher
// in my cs 638 class that I had in the Fall of 1999
float floor1[4] = { 0.5f, 0.5f, 0.5f, 1 };
float floor2[4] = { 0.4f, 0.4f, 0.4f, 1 };
void animWindow::drawFloor( float floorHeight )
{
  // parameters:
  int xSquares = 20;
  int ySquares = 20;
  float maxX = 200, maxY = 200;
  float minX = -200, minY = -200;

	glDisable(GL_LIGHTING);		// make the colors absolute
	glEnable(GL_DEPTH_TEST);	// we actually need this turned on
								// or else it isn't opaque

	int x,y,v[3],i;
	float xp,yp,xd,yd;
	v[2] = 0;
	xd = (maxX - minX) / ((float) xSquares);
	yd = (maxY - minY) / ((float) ySquares);
	glBegin(GL_QUADS);
	for(x=0,xp=minX; x<xSquares; x++,xp+=xd)
		for(y=0,yp=minY,i=x; y<ySquares; y++,i++,yp+=yd) {
			glColor4fv(i&1 ? floor1:floor2);
			glNormal3f(maxX/1,1,maxY/2); 
			glVertex3d(xp,floorHeight,yp);
			glVertex3d(xp,floorHeight,yp+yd);
			glVertex3d(xp+xd,floorHeight,yp+yd);
			glVertex3d(xp+xd,floorHeight,yp);
		}
	glEnd();
}

// handle whatever events come our way
//
// the mouse behavior is:
// right mouse = control view. Basically, it orbits around the center
// (if we're looking in 3D)
// left mouse = select object.  
// we'll remember where the mouse goes down for later
static int mouseDownX, mouseDownY;
// remember where the mouse was last when we last processed the mouseEvent
static int lastX, lastY;
// where the eye started to move things around
int animWindow::handle(int e)
{
	switch(e) {
		case FL_SHOW:		// you must handle this, or not be seen!
			show();
			return 1;
		case FL_PUSH:
			lastX = mouseDownX = Fl::event_x();	// remeber where the mouse went down
			lastY = mouseDownY = Fl::event_y();
			switch(Fl::event_button()) 
			{
				case 1:
					lookAtXOld = lookAtX ;		// In all cases, upon the initial click, store
					lookAtZOld = lookAtZ ;		// the old global values as old values.
					rotateCXOld = rotateCX ;
					rotateCYOld = rotateCY ;
					lookAtYOld = lookAtY ;
					break ;
				case 2:
					lookAtXOld = lookAtX ;
					lookAtZOld = lookAtZ ;
					lookAtYOld = lookAtY ;
					break ;
				case 3:
					lookAtXOld = lookAtX ;
					lookAtZOld = lookAtZ ;
					scaleCOld = scaleC ;
					break;
			};
			return 1;
		case FL_DRAG: // if the user drags the mouse
			switch(Fl::event_button()) 
			{
				case 1:
					if (Fl::event_state(FL_CAPS_LOCK))
					{
						// Calculate the camera's vector back to the center-of-interest.
						float cameraX = (float)( cos((rotateCY * PI)/180) * scaleC * cos((rotateCX * PI)/180) ) ;
						float cameraZ = (float)( sin((rotateCY * PI)/180) * scaleC * cos((rotateCX * PI)/180) ) ;

						// Assign translate X and Z to be the normal vec of the to-the-center-of-interest
						// This disregards the Y vector, as you are really not paying attention to that anyways.
						translateX = cameraX ;
						translateZ = cameraZ ;

						// Normalize yer' translate vector...
						float newStinkinX = translateX ;
						translateX = (float)(translateX / sqrt((translateX * translateX) + (translateZ * translateZ))) ;		//Y is guaranteed to be 0
						translateZ = (float)(translateZ / sqrt((newStinkinX * newStinkinX) + (translateZ * translateZ))) ;		//Y is guaranteed to be 0

						// Effectively pan the camera depending on which direction you drag...
						lookAtX = lookAtXOld + -((Fl::event_x() - mouseDownX) * translateX) ;
						lookAtZ = lookAtZOld + -((Fl::event_x() - mouseDownX) * translateZ) ;
						lookAtY = lookAtYOld + -(Fl::event_y() - mouseDownY) ;
					}
					else
					{
						// If caps lock isn't down, just rotate the camera around its center of interest.
						rotateCX = rotateCXOld + Fl::event_y() - mouseDownY ;
						rotateCY = rotateCYOld + Fl::event_x() - mouseDownX ;
					}
					damage(1) ;
					break ;
				case 2:
					{
						// Calculate the camera's vector back to the center-of-interest.
						float cameraX = (float)( cos((rotateCY * PI)/180) * scaleC * cos((rotateCX * PI)/180) ) ;
						float cameraZ = (float)( sin((rotateCY * PI)/180) * scaleC * cos((rotateCX * PI)/180) ) ;

						// Assign translate X and Z to be the normal vec of the to-the-center-of-interest
						// This disregards the Y vector, as you are really not paying attention to that anyways.
						translateX = cameraX ;
						translateZ = cameraZ ;

						// Normalize yer' lookats...
						float newStinkinX = translateX ;
						translateX = (float)(translateX / sqrt((translateX * translateX) + (translateZ * translateZ))) ;		//Y is guaranteed to be 0
						translateZ = (float)(translateZ / sqrt((newStinkinX * newStinkinX) + (translateZ * translateZ))) ;		//Y is guaranteed to be 0
	
						// Effectively pan the camera depending on which direction you drag...
						lookAtX = lookAtXOld + -((Fl::event_x() - mouseDownX) * translateX) ;
						lookAtZ = lookAtZOld + -((Fl::event_x() - mouseDownX) * translateZ) ;
						lookAtY = lookAtYOld + -(Fl::event_y() - mouseDownY) ;
						damage(1);
					}
					break ;
				case 3:	
					if (Fl::event_state(FL_CAPS_LOCK))
					{
						// Calculate the camera's vector back to the center-of-interest.
						float cameraX = -1 * (float)( cos((rotateCY * PI)/180) * scaleC * cos((rotateCX * PI)/180) ) ;
						float cameraZ = -1 * (float)( sin((rotateCY * PI)/180) * scaleC * cos((rotateCX * PI)/180) ) ;

						// Assign translate X and Z to be the normal vec of the to-the-center-of-interest
						// This disregards the Y vector, as you are really not paying attention to that anyways.
						translateX = cameraZ ;
						translateZ = -cameraX ;

						// Normalize yer' lookats...
						float newStinkinX = translateX ;
						translateX = (float)(translateX / sqrt((translateX * translateX) + (translateZ * translateZ))) ;		//Y is guaranteed to be 0
						translateZ = (float)(translateZ / sqrt((newStinkinX * newStinkinX) + (translateZ * translateZ))) ;		//Y is guaranteed to be 0

						// Effectively pan the camera FORWARD depending on which direction you drag...
						lookAtX = lookAtXOld + -((Fl::event_x() - mouseDownX) * translateX) ;
						lookAtZ = lookAtZOld + -((Fl::event_x() - mouseDownX) * translateZ) ;
					}
					else
					{
						// If Caps Lock isn't down, just scale.
						scaleC = scaleCOld + Fl::event_x() - mouseDownX ;
					}
					damage(1);
					break ;
			}
		default:
			return 0;
	}
}

