/********************************************************************************************
*	Files: graphEditor.h and graphEditor.cpp
*	Written by Steve Dutcher and Andy Gardner
*
*	Some key things to note:
*		This class contains lists of motion curves locally, and saves any global manipulation
*		only until the data has been manipulated.
*		This insures that the graph editor will not, under any strange circumstance, destroy 
*		the global data of the joint objects.  Interaction with the data is done on a once-in and
*		once-out basis and stored accordingly.
*
*		The data structure for storing the x y and z rotations for a point is as follows:
*		Each joint has a vector of points, this vector is arranged in the following structure
*		<(keyframeOne)RX RY RZ TX TY TZ (keyframeTwo)RX RY RZ TX TY TZ ... >
*		This is very important to keep straight, as any time you would like to check the second
*		RX value on a curve, it is necessary to add SIX (accounted for by a const int) to the
*		Vector offset.
*
*		Everything else should be straightforward.
*
*	Read the graphEditor.h file for descriptions of each function if not given
*********************************************************************************************/
#include <stdio.h>
#include "RunButton.h"
#include "graphEditor.h"
#include "3DUtils.h"

/****************************************************************************
*
*  This secion is for the Graph Editor Window which hold the 
*  graph and all the buttons
*
*****************************************************************************/

enum names {RX, RY, RZ, TX, TY, TZ};

void Graph_Editor_Window_cb(Fl_Widget*, void*);

// For any buttons create an int and pass that to their callback
// Then in the funciton Graph_Editor_Window_CB(int action) you compare
// action with these numbers to see which button was pressed
const unsigned int EXITBUTTON = 1001;

void rx_button_cb(Fl_Widget *myself, void *gw){
	((Gl_Graph_Window*)gw)->setShown( RX, ((Fl_Button*)myself)->value() );
}
void ry_button_cb(Fl_Widget *myself, void *gw){
	((Gl_Graph_Window*)gw)->setShown( RY, ((Fl_Button*)myself)->value() );
}
void rz_button_cb(Fl_Widget *myself, void *gw){
	((Gl_Graph_Window*)gw)->setShown( RZ, ((Fl_Button*)myself)->value() );
}
void tx_button_cb(Fl_Widget *myself, void *gw){
	((Gl_Graph_Window*)gw)->setShown( TX, ((Fl_Button*)myself)->value() );
}
void ty_button_cb(Fl_Widget *myself, void *gw){
	((Gl_Graph_Window*)gw)->setShown( TY, ((Fl_Button*)myself)->value() );
}
void tz_button_cb(Fl_Widget *myself, void *gw){
	((Gl_Graph_Window*)gw)->setShown( TZ, ((Fl_Button*)myself)->value() );
}
void interp_button_cb(Fl_Widget *myself, void *gw){
	((Gl_Graph_Window*)gw)->interpBetweenSelected();
}

// Create and initialize all graph editor buttons and windows.
// This is called once upon inception and executes many fltk commands.
Graph_Editor_Window::Graph_Editor_Window(char* name, motionData *md, Fl_Slider* slide) :
	Fl_Window(0, 0, 700, 360, name),
	jBrowser(10, 50, 200, h()-60, md, (Fl_Widget*)this, "Joint Selector"),
	graphWindow(220, 10, w()-230, h()-65, md, slide, (Fl_Browser*)&jBrowser),
	rxButton(220, 310, 110, 20, "Rotate X"),
	ryButton(340, 310, 110, 20, "Rotate Y"),
	rzButton(460, 310, 110, 20, "Rotate Z"),
	txButton(220, 335, 110, 20, "Translate X"),
	tyButton(340, 335, 110, 20, "Translate Y"),
	tzButton(460, 335, 110, 20, "Translate Z"),
	interpButton(580, 310, 110,45, "Interpolate"),
	exitButton(10, 10, 200, 30, "Exit Button")
{
	rxButton.selection_color( FL_RED );
	ryButton.selection_color( FL_GREEN );
	rzButton.selection_color( FL_BLUE );
	txButton.selection_color( FL_YELLOW );
	tyButton.selection_color( FL_MAGENTA );
	tzButton.selection_color( FL_CYAN );

	rxButton.callback( (Fl_Callback*)rx_button_cb, &graphWindow );	
	ryButton.callback( (Fl_Callback*)ry_button_cb, &graphWindow );	
	rzButton.callback( (Fl_Callback*)rz_button_cb, &graphWindow );	
	txButton.callback( (Fl_Callback*)tx_button_cb, &graphWindow );	
	tyButton.callback( (Fl_Callback*)ty_button_cb, &graphWindow );	
	tzButton.callback( (Fl_Callback*)tz_button_cb, &graphWindow );
	interpButton.callback( (Fl_Callback*)interp_button_cb, &graphWindow );
	
	rxButton.value( 1 );
	ryButton.value( 1 );
	rzButton.value( 1 );
	txButton.value( 1 );
	tyButton.value( 1 );
	tzButton.value( 1 );
	
	group = new Fl_Group(210, 0, 490, 300 );
		mdata = md;
		exitButton.callback( (Fl_Callback*)Graph_Editor_Window_cb, (void*)EXITBUTTON );
	//	group->add(jBrowser);
		group->add(graphWindow);
		resizable(group);
		show();
		end();
		size_range(0,0,2000,2000);	
}

// Exits the graph editor window.
void Graph_Editor_Window::Graph_Editor_Window_CB(int action)
{
	if( action == EXITBUTTON )
	{
		delete this;
	}	
}

// A graph editor window callback to handle fltk windowing routines.
void Graph_Editor_Window_cb( Fl_Widget *w, void* action)
{
	((Graph_Editor_Window*)(w->parent()))->Graph_Editor_Window_CB( (int)action );	
}





/****************************************************************************
*
*  This secion is for the GL_Window  a.k.a. the graph
*
*****************************************************************************/

Gl_Graph_Window::Gl_Graph_Window(int x, int y, int w, int h, motionData* md, Fl_Slider* slide, Fl_Browser* browse) :
	Fl_Gl_Window(x, y, w, h)
{
	for(int i=0; i<6; i++)
		whoIsShown[i] = 1;

	// Initialize all the necessary variables.

	mdata = md ;
	left = bottom = -8.0f;
	right = top = 8.0f;

	bigNameString = NULL ;
	numNodeNames  = 0 ; 

	time = slide ;
	jointBrowser = browse ;

	((RunSlider*)time)->addWidgetToDamageList( this );
}

// Destroys the graph window when necessary.
Gl_Graph_Window::~Gl_Graph_Window(){
	((RunSlider*)time)->removeWidgetFromDamageList( this );
	delete[] bigNameString ;
}

// Sets whether or not the current graph editor is shown or not.
void Gl_Graph_Window::setShown(int button, int value)
{
	whoIsShown[button] = value;
	damage(1);
}

// Keeps track of where the mouse was initially pressed and last clicked.
static int mouseDownX, mouseDownY, lastX, lastY ;

// Flag turned on only if nothing is being dragged.
static int noDragFlag ;

// Due to the nature of the point vectors (see .h file), incrementing one keyframe
// along the list actually increments the joint pointer by 6 values.  This takes that
// into account
static const int ONE = 6 ;

// Handle what FLTK stuff comes our way.
int Gl_Graph_Window::handle(int event)
{
	float ratioWidth  = 0 ;
	float ratioHeight = 0 ;

	switch(event) 
	{
		case FL_KEYBOARD:
			if (Fl::event_key() == FL_Delete)
			{
				if (selectionFlag == true)
					interpBetweenSelected() ;	// The delete key interpolates as well as the
			}									// interpolate button.
		return 1 ;
		case FL_PUSH:
			// mouse down event ...
			// position in Fl::event_x() and Fl::event_y()
			lastX = mouseDownX = Fl::event_x() ;
			lastY = mouseDownY = Fl::event_y() ;
			if (Fl::event_button() == 1)		// Set everything ready to start drawing the
			{									// selection box.
				boxFlag = false ;
				clickX = (float)lastX ;
				clickY = (float)lastY ;
				ratioWidth  = clickX / w() ;
				ratioHeight = clickY / h() ;
				clickX = ((right - left) * ratioWidth) + left ;
				clickY = -1 * (((top - bottom) * ratioHeight) - top) ;
				damage(1)  ;
			}
			if (Fl::event_button() == 2)		// Frame all of the curves in the window.
			{
				frameAll() ;
				damage(1)  ;
			}
			if (Fl::event_button() == 3)		// If you're not dragging, you are now.
			{
				if (noDragFlag == 1)
					noDragFlag = 0 ;
			}
		return 1;
		case FL_DRAG:
			// mouse moved while down event ...
			switch (Fl::event_button())
			{
				case 1:
					// If caps lock is on, the left mouse button scales the graph editor window.
					if (Fl::event_state(FL_CAPS_LOCK))
					{						
						right  -= (Fl::event_x() - lastX) * ((right - left) / 120) ;
						left   += (Fl::event_x() - lastX) * ((right - left) / 120) ;
						top    -= (Fl::event_y() - lastY) * ((top - bottom) / 120) ;
						bottom += (Fl::event_y() - lastY) * ((top - bottom) / 120) ;
						lastX = Fl::event_x() ;
						lastY = Fl::event_y() ;
					}
					else	// If caps lock is not down, this creates the selection box.
					{
						dragX = lastX = Fl::event_x() ;
						dragY = lastY = Fl::event_y() ;

						boxFlag = true ;

						dragX = (float)lastX ;
						dragY = (float)lastY ;
						ratioWidth  = dragX / w() ;
						ratioHeight = dragY / h() ;
						dragX = ((right - left) * ratioWidth) + left ;
						dragY = -1 * (((top - bottom) * ratioHeight) - top) ;
					}
					damage (1) ;
					break ;
				case 3:	
					// If caps lock is on, the right mouse button translates the graph editor
					if (Fl::event_state(FL_CAPS_LOCK))
					{
						noDragFlag = 1 ;
						right  -= (Fl::event_x() - lastX) * ((right - left) / 500) ;		//This is the pan code...
						left   -= (Fl::event_x() - lastX) * ((right - left) / 500) ;
						top    += (Fl::event_y() - lastY) * ((right - left) / 500) ;
						bottom += (Fl::event_y() - lastY) * ((right - left) / 500) ;
						lastX = Fl::event_x() ;
						lastY = Fl::event_y() ;
					}
					else	// Otherwise it drags the selected points.
					{
						dragPoints( -((Fl::event_y() - lastY) * ((top - bottom) / 90.0f)) ) ;
						lastY = Fl::event_y() ;
					}
					damage(1) ;
					break ;
			}
		return 1;
		case FL_RELEASE:    
			// mouse up event ...
			if (Fl::event_button() == 1)
			{
				// If you let go of the left button, this calculates what is selected.
				for (int i = 0; i < lineVector.size(); i += 2)
				{
					pVector tempPointVec = lineVector[i]   ;
					pVector tempSelMask  = lineVector[i+1] ;

					//Clear the selection mask first...
					for (int j = 0; j < tempSelMask.size(); j++)
					{
						tempSelMask[j] = 0.0f ;
					}

					//If you didn't do a mouse click in one point only (did you drag)...
					if ((dragX != 0.0f) && (dragY != 0.0f))
					{
						int tmpFlag = 0 ;
						//printf ("Line x %d\n", i) ;
						if( whoIsShown[RX] == 1 ){
							lineSelectionProcess(tempPointVec, tempSelMask, 0) ;
							if (selectionFlag == true) tmpFlag = 1 ;
							//printf ("Line y %d\n", i) ;
						}
						if( whoIsShown[RY] == 1 ){
							lineSelectionProcess(tempPointVec, tempSelMask, 1) ;
							if (selectionFlag == true) tmpFlag = 1 ;
							//printf ("Line z %d\n", i) ;
						}
						if( whoIsShown[RZ] == 1 ){
							lineSelectionProcess(tempPointVec, tempSelMask, 2) ;
							if (selectionFlag == true) tmpFlag = 1 ;
						}
						if (tmpFlag == 1) selectionFlag = true ;
					}

					lineVector[i+1] = tempSelMask ;
				}
					
				clickX = clickY = dragX = dragY = 0 ;
				boxFlag = false ;
				damage (1) ;
			}
			if (Fl::event_button() == 3)	//Now that you're done dragging the points...
			{
				if (noDragFlag == 0)
					doneDraggingPoints() ;
				else
					noDragFlag = 1 ;
			}

		return 1;
		case FL_FOCUS :
		case FL_UNFOCUS :
			// Return 1 if you want keyboard events, 0 otherwise
		return 1;
		case FL_SHORTCUT:
			// shortcut, key is in Fl::event_key(), ascii in Fl::event_text()
			// Return 1 if you understand/use the shortcut event, 0 otherwise...
		return 1;
		default:
			// pass other events to the base class...
		return Fl_Gl_Window::handle(event);
	}
}

// This function is called either externally (dragFlag = 0) or internally (dragFlag = 1)
// It gathers a space-delimited list of names (selection) and the number of names to expect
// in that list (numNames) and creates curves accordingly.
void Gl_Graph_Window::initGraphEditor(char * selection, int numNames, int dragFlag)
{
	// Init current vectors of data.
	lineVector.clear() ;
	isEffector.clear() ;

	if (selection == NULL)
		selection = strdup (bigNameString) ;

	topMax = bottomMax = 0 ;

	if (bigNameString != NULL)
		delete[] bigNameString ;

	// Init classes variables according to what was passed in.
	numNodeNames = numNames ;
	bigNameString = strdup (selection) ;

	for (int j = 0; j < numNames; j++)
	{
		float xx, xy, xz ;
		pVector pointVector ;
		pVector selectionMask ;
		char tempString[1000] ;

		joint* j ;

		// Parse through the string.
		sscanf (selection, "%s", tempString) ;
		selection = selection + strlen(tempString) + 1 ;

		//printf ("Initializing for %s\n", tempString) ;

		j = mdata->getJoint(tempString) ;

		//Push yourself what is an effector and what isn't...
		if (j->getType() == EFFECTOR)
		{
			//printf("Pushing 1 back for Effector %s\n", tempString) ;
			isEffector.push_back(1) ;
			isEffector.push_back(0) ;		//Accomodate for the selectionMask
		}
		else
		{
			//printf ("Pushing 0 back for nonE \n") ;
			isEffector.push_back(0) ;
			isEffector.push_back(0) ;		//Accomodate for the selectionMask
		}

		for (int i = 1; i <= mdata->getNumberOfFrames(); i++)
		{
			if (j->getType() != EFFECTOR)
			{
				//Get the rotations and store 'em in the first 3 slots for the point vector...
				j->getRotation(i, xx, xy, xz) ;
				pointVector.push_back(xx) ;
				pointVector.push_back(xy) ;
				pointVector.push_back(xz) ;

				// Set the top and bottom maximums if necessary.
				if (xx > topMax) topMax = xx ;
				else if (xx < bottomMax) bottomMax = xx ;
				if (xy > topMax) topMax = xy ;
				else if (xy < bottomMax) bottomMax = xy ;
				if (xz > topMax) topMax = xz ;
				else if (xz < bottomMax) bottomMax = xz ;
			}
			else
			{
				//If you're an end effector Buffer those rotations with 0's...
				pointVector.push_back(0.0f) ;
				pointVector.push_back(0.0f) ;
				pointVector.push_back(0.0f) ;
			}

			//Get the translations and store 'em in the second 3 slots for the point vector...
			j->getEulerMarkerPosition(i, xx, xy, xz) ;
			pointVector.push_back(xx) ;
			pointVector.push_back(xy) ;
			pointVector.push_back(xz) ;

			// Set the top and bottom maximums if necessary.
			if (xx > topMax) topMax = xx ;
			else if (xx < bottomMax) bottomMax = xx ;
			if (xy > topMax) topMax = xy ;
			else if (xy < bottomMax) bottomMax = xy ;
			if (xz > topMax) topMax = xz ;
			else if (xz < bottomMax) bottomMax = xz ;

			//printf ("%f %f %f\n", rx, ry, rz) ;

			//Push six 0's for the selection mask for this pointVector...
			selectionMask.push_back(0.0f) ;selectionMask.push_back(0.0f) ;selectionMask.push_back(0.0f) ;
			selectionMask.push_back(0.0f) ;selectionMask.push_back(0.0f) ;selectionMask.push_back(0.0f) ;
		}
		// Add the vectors to the list.
		lineVector.push_back(pointVector)   ;
		lineVector.push_back(selectionMask) ;
	}

	if (!dragFlag)
		frameAll() ;						

	//printf ("%f %f %f %f\n", left, right, bottom, top) ;
	activate () ;
	damage (1) ;												//These do not work - why?
	damage (FL_DAMAGE_ALL) ;
	redraw() ;
	//draw() ;
}

// FLTK's draw function for drawing the GL portion of the graph window.
void Gl_Graph_Window::draw()
{
	glClearColor(0,0,0,0);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glMatrixMode(GL_PROJECTION);

	glViewport(0,0,w(),h());

	glPushMatrix()  ;
	glLoadIdentity();

	glOrtho(left, right, bottom, top, -8.0, 8.0);
	glMatrixMode(GL_MODELVIEW);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	gl_font(FL_COURIER + FL_BOLD, 10) ;

	//For drawing the side and bottom numbers...
	for (int m = 0; m < 10; m++)
	{
		char numberTemp[1000] ;

		//This draws the horizontal lines...
		glColor4f(0.23f, 0.23f, 0.23f, 1.0f) ;
		glBegin(GL_LINE_STRIP);
			glVertex2f(left-2,  ( top - (m * ((top-bottom) / 10))) ) ;
			glVertex2f(right+2, ( top - (m * ((top-bottom) / 10))) ) ;
		glEnd() ;

		//This is for the left side of the window...
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f) ;
		int x = (int)( top - (((top-bottom) / 10 ) * m ))  ;

		sprintf(numberTemp, "%d", x) ;
		gl_draw(numberTemp, (float)(left + ((right-left)/50)), (float)( top - (m * ((top-bottom) / 10))) ) ;

		//This draws the vertical lines...
		glColor4f(0.23f, 0.23f, 0.23f, 1.0f) ;
		glBegin(GL_LINE_STRIP);
			glVertex2f(( right - (m * ((right-left) / 10))), bottom-2) ;
			glVertex2f(( right - (m * ((right-left) / 10))), top+2) ;
		glEnd() ;

		//This is for the bottom of the window...
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f) ;
		x = (int)( right - (((right-left) / 10 ) * m ))  ;

		sprintf(numberTemp, "%d", x) ;
		gl_draw(numberTemp, (float)( right - (m * ((right-left) / 10))), (float)(bottom + ((top-bottom)/50)) ) ;

	}

	//For drawing the time slider line...
	glBegin(GL_LINE_STRIP);
		glColor4f(1.0f, 1.0f, 0.0f, 1.0f) ;
		glVertex2f((int)time->value(), top + 2) ;
		glVertex2f((int)time->value(), bottom - 2) ;
	glEnd();

	//Here's where we look at what the values of the joint browser are...
	if (((JointBrowser*)jointBrowser)->didIChange())
	{
		char temp [1000];
		int numNames = ((JointBrowser*)jointBrowser)->getSelectedNames(temp) ;

		//printf ("asdf %d\n", numNames) ;
		if (numNames > 0) 
		{
			//printf ("This is the string i get 1 %s\n", temp) ;
			initGraphEditor(temp, numNames, 0) ;
		}
	}

	//Here's what actually draws the lines...
	for (int i = 0; i < lineVector.size(); i += 2)
	{
		float time = 1.0f ;
		pVector tempPointVec = lineVector[i]   ;
		pVector tempSelMask  = lineVector[i+1] ;

		if (isEffector[i] == 0)
		{
		    if( whoIsShown[RX] == 1 )
			//The rX line...
			for (int j = 0; j < (tempPointVec.size() - ONE); j += ONE)
			{
				glColor4f(1.0f, 0.0f, 0.0f, 1.0f) ;
				glBegin (GL_LINES);
					glVertex2f(time, tempPointVec[j]);
					glVertex2f(time+1, tempPointVec[j+ONE]);
				glEnd() ;

				if (tempSelMask[j] == 0.0f)
					glColor4f(1.0f, 1.0f, 1.0f, 0.5f) ;
				else
					glColor4f(1.0f, 1.0f, 0.0f, 0.9f) ;

				//Dot drawing.
				drawBox(time, tempPointVec[j], 0.0f, (float)((right-left) / 85), (float)((top-bottom) / 70) ) ;

				//Draw that last dot - off by one fix...
				if (tempPointVec.size() - ONE - 6 == j)
				{
					if (tempSelMask[tempSelMask.size() - 6] == 0.0f)
						glColor4f(1.0f, 1.0f, 1.0f, 0.5f) ;
					else
						glColor4f(1.0f, 1.0f, 0.0f, 0.9f) ;

					drawBox(time+1, tempPointVec[j+ONE], 0.0f, (float)((right-left) / 85), (float)((top-bottom) / 70) ) ;
				}

				time += 1 ;
			}

			time = 1 ;

		    if( whoIsShown[RY] == 1 )
			//The rY line...
			for (int k = 1; k < (tempPointVec.size() - ONE); k += ONE)
			{
				glColor4f(0.0f, 1.0f, 0.0f, 1.0f) ;
				glBegin (GL_LINES);
					glVertex2f(time, tempPointVec[k]);
					glVertex2f(time+1, tempPointVec[k+ONE]);
				glEnd() ;

				if (tempSelMask[k] == 0.0f)
					glColor4f(1.0f, 1.0f, 1.0f, 0.5f) ;
				else
					glColor4f(1.0f, 1.0f, 0.0f, 0.9f) ;

				//Dot drawing.
				drawBox(time, tempPointVec[k], 0.0f, (float)((right-left) / 85), (float)((top-bottom) / 70) ) ;

				//Draw that last dot - off by one fix...
				if (tempPointVec.size() - ONE - 5 == k)
				{
					if (tempSelMask[tempSelMask.size() - 5] == 0.0f)
						glColor4f(1.0f, 1.0f, 1.0f, 0.5f) ;
					else
						glColor4f(1.0f, 1.0f, 0.0f, 0.9f) ;

					drawBox(time+1, tempPointVec[k+ONE], 0.0f, (float)((right-left) / 85), (float)((top-bottom) / 70) ) ;
				}

				time += 1 ;
			}


			time = 1 ;

		    if( whoIsShown[RZ] == 1 )
			//The rZ line...
			for (int l = 2; l < (tempPointVec.size() - ONE); l += ONE)
			{
				glColor4f(0.0f, 0.0f, 1.0f, 1.0f) ;
				glBegin (GL_LINES);
					glVertex2f(time, tempPointVec[l]);
					glVertex2f(time+1, tempPointVec[l+ONE]);
				glEnd() ;

				if (tempSelMask[l] == 0.0f)
					glColor4f(1.0f, 1.0f, 1.0f, 0.5f) ;
				else
					glColor4f(1.0f, 1.0f, 0.0f, 0.9f) ;

				//Dot drawing.
				drawBox(time, tempPointVec[l], 0.0f, (float)((right-left) / 85), (float)((top-bottom) / 70) ) ;

				//Draw that last dot - off by one fix...
				if (tempPointVec.size() - ONE - 4 == l)
				{
					if (tempSelMask[tempSelMask.size() - 4] == 0.0f)
						glColor4f(1.0f, 1.0f, 1.0f, 0.5f) ;
					else
						glColor4f(1.0f, 1.0f, 0.0f, 0.9f) ;

					drawBox(time+1, tempPointVec[l+ONE], 0.0f, (float)((right-left) / 85), (float)((top-bottom) / 70) ) ;
				}

				time += 1 ;
			}
		}

		time = 1 ;
	   
	   if( whoIsShown[TX] == 1 )
		//The tX line...
		for (int m = 3; m < (tempPointVec.size() - ONE); m += ONE)
		{
			glColor4f(1.0f, 1.0f, 0.0f, 1.0f) ;
			glBegin (GL_LINES);
				glVertex2f(time, tempPointVec[m]);
				glVertex2f(time+1, tempPointVec[m+ONE]);
			glEnd() ;

			time += 1 ;
		}


		time = 1 ;

	   if( whoIsShown[TY] == 1 )
		//The tY line...
		for (int n = 4; n < (tempPointVec.size() - ONE); n += ONE)
		{
			glColor4f(1.0f, 0.0f, 1.0f, 1.0f) ;
			glBegin (GL_LINES);
				glVertex2f(time, tempPointVec[n]);
				glVertex2f(time+1, tempPointVec[n+ONE]);
			glEnd() ;

			time += 1 ;
		}


		time = 1 ;

	   if( whoIsShown[TZ] == 1 )
		//The tZ line...
		for (int p = 5; p < (tempPointVec.size() - ONE); p += ONE)
		{
			glColor4f(0.0f, 1.0f, 1.0f, 1.0f) ;
			glBegin (GL_LINES);
				glVertex2f(time, tempPointVec[p]);
				glVertex2f(time+1, tempPointVec[p+ONE]);
			glEnd() ;

			time += 1 ;
		}
	}

	// This draws the selection box if that flag is turned on.
	if (boxFlag)
	{
		glColor4f(0.8f, 0.8f, 0.8f, 0.25f) ;
		glRectf(clickX, clickY, dragX, dragY) ;
		glColor4f(0.8f, 0.8f, 0.8f, 1.0f) ;
		glBegin(GL_LINE_LOOP) ;
			glVertex2d(clickX, clickY) ;
			glVertex2d(clickX, dragY) ;
			glVertex2d(dragX, dragY) ;
			glVertex2d(dragX, clickY) ;
		glEnd() ;
	}

	//Pop!
	glPopMatrix() ;
}

// Frames the data in the graph window according to the class' max and min data values.
void Gl_Graph_Window::frameAll()
{
	right  = mdata->getNumberOfFrames() + ((right - left) * 0.1f);
	left   = -((right - left) * 0.1f) ;
	top    = topMax + ((right - left) * 0.1f) ;
	bottom = bottomMax - ((right - left) * 0.1f) ;
}

// An internal function used to select, according to an 'offset' the proper points dictated
// by the current selection box.  This is mainly done to save a lot of extra code.
void Gl_Graph_Window::lineSelectionProcess(pVector &tempPointVec, pVector &tempSelMask, int offset)
{
	float time = 1.0f ;

	selectionFlag = false ;

	int selFlag = 0 ;

	//For the X-line selection...
	for (int j = offset; j < tempPointVec.size(); j += ONE) {
		if (dragX > clickX) {  //Drag RIGHT...
			if ((time > clickX) && (time < dragX)) {
				if (dragY > clickY)	{  //Drag UP...
					if ((tempPointVec[j] > clickY) && (tempPointVec[j] < dragY)) {
						//printf ("Got a point right-up %f line \n", tempPointVec[j]) ;
						tempSelMask[j] = 1.0f ;
						selectionFlag = true ;
					}
				}
				else {  //Drag DOWN...
					if ((tempPointVec[j] < clickY) && (tempPointVec[j] > dragY)) {
						//printf ("Got a point right-down %f line \n", tempPointVec[j]) ;
						tempSelMask[j] = 1.0f ;
						selectionFlag = true ;
					}
				}
			}
		}
		else {  //Drag LEFT...
			if ((time < clickX) && (time > dragX)) {
				if (dragY > clickY)	{  //Drag UP...
					if ((tempPointVec[j] > clickY) && (tempPointVec[j] < dragY)) {
						//printf ("Got a point left-up %f line \n", tempPointVec[j]) ;
						tempSelMask[j] = 1.0f ;
						selectionFlag = true ;
					}
				}
				else {  //Drag DOWN...
					if ((tempPointVec[j] < clickY) && (tempPointVec[j] > dragY))
					{
						//printf ("Got a point left-down %f line \n", tempPointVec[j]) ;
						tempSelMask[j] = 1.0f ;
						selectionFlag = true ;
					}
				}
			}
		}
		time += 1.0f ;
	}
}

// Interpolates all of the selected points in the graph editor.  No values passed in or out.
void Gl_Graph_Window::interpBetweenSelected(void)
{
	char tempString[1000]   ;
	joint* jointPtr ;
	char * selection = bigNameString ;

	int beginningPointX = 0 ;
	int endPointX       = 0 ;
	int deleteFlagX     = 0 ;
	int beginningPointY = 0 ;
	int endPointY       = 0 ;
	int deleteFlagY     = 0 ;
	int beginningPointZ = 0 ;
	int endPointZ       = 0 ;
	int deleteFlagZ     = 0 ;

	int startQ, stopQ;
	stopQ = -1;
	startQ = 32000;

	for (int i = 0; i < lineVector.size(); i += 2)
	{
		pVector tempPointVec = lineVector[i]   ;
		pVector tempSelMask  = lineVector[i+1] ;

		sscanf (selection, "%s", tempString) ;
		selection = selection + strlen(tempString) + 1 ;

		//printf ("asd %s\n", tempString) ;

		int doFlag = 0 ;

		if (isEffector[i] == 0)
		{
			for (int j = 0; j < tempPointVec.size(); j += ONE)
			{
				//FIRST we figure out what is and isn't selected.

				if (tempSelMask[j] == 1.0f && deleteFlagX == 0){
					beginningPointX = (j/ONE) ;						//This actually accounts for element 0
					deleteFlagX = 1 ;								//  really being element 1...
				}
				if (tempSelMask[j] == 0.0f && deleteFlagX == 1){
					endPointX = (j/ONE)+1 ;
					deleteFlagX = 0 ;
					doFlag = 1 ;
				}
				else if (tempSelMask[j] == 1.0f && deleteFlagX == 1 && (j/ONE+1) == mdata->getNumberOfFrames())
				{
					endPointX = (j/ONE)+1 ;
					deleteFlagX = 0 ;
					doFlag = 1 ;
				}	//Accounts for last frame being selected...

				if (tempSelMask[j+1] == 1.0f && deleteFlagY == 0){
					beginningPointY = (j/ONE) ;
					deleteFlagY = 1 ;
				}
				if (tempSelMask[j+1] == 0.0f && deleteFlagY == 1){
					endPointY = (j/ONE)+1 ;
					deleteFlagY = 0 ;
					doFlag = 1 ;
				}
				else if (tempSelMask[j+1] == 1.0f && deleteFlagY == 1 && (j/ONE+1) == mdata->getNumberOfFrames())
				{
					endPointY = (j/ONE)+1 ;
					deleteFlagY = 0 ;
					doFlag = 1 ;
				}	//Accounts for last frame being selected...

				if (tempSelMask[j+2] == 1.0f && deleteFlagZ == 0){
					beginningPointZ = (j/ONE) ;
					deleteFlagZ = 1 ;
				}
				if (tempSelMask[j+2] == 0.0f && deleteFlagZ == 1){
					endPointZ = (j/ONE)+1 ;
					deleteFlagZ = 0 ;
					doFlag = 1 ;
				}
				else if (tempSelMask[j+2] == 1.0f && deleteFlagZ == 1 && (j/ONE+1) == mdata->getNumberOfFrames())
				{
					endPointZ = (j/ONE)+1 ;
					deleteFlagZ = 0 ;
					doFlag = 1 ;
				}	//Accounts for last frame being selected...

				//SECOND, if something is selected, and we have a range, interpolate the data.

				if (doFlag == 1)
				{
					jointPtr = mdata->getJoint(tempString) ;

					if (endPointX != 0)
					{
						//printf ("BeginningPoint X : %d  EndPoint X : %d    %s\n", beginningPointX, endPointX, tempString) ;
						jointPtr->interpolateRX(beginningPointX, endPointX) ;
//						jointPtr->interpolateQuaternionRX(beginningPointX, endPointX) ;
					}

					if (endPointY != 0)
					{
						//printf ("BeginningPoint Y : %d  EndPoint Y : %d    %s\n", beginningPointY, endPointY, tempString) ;
						jointPtr->interpolateRY(beginningPointY, endPointY) ;
//						jointPtr->interpolateQuaternionRY(beginningPointY, endPointY) ;
					}

					if (endPointZ != 0)
					{
						//printf ("BeginningPoint Z : %d  EndPoint Z : %d    %s\n", beginningPointZ, endPointZ, tempString) ;
						jointPtr->interpolateRZ(beginningPointZ, endPointZ) ;
//						jointPtr->interpolateQuaternionRZ(beginningPointZ, endPointZ) ;
					}

					if( (beginningPointX != endPointX) && (beginningPointX < startQ) )
						startQ = beginningPointX;
					if( (beginningPointY != endPointY) && (beginningPointY < startQ) )
						startQ = beginningPointY;
					if( (beginningPointZ != endPointZ) && (beginningPointZ < startQ) )
						startQ = beginningPointZ;
					if( (endPointX != 0) && (endPointX > stopQ) )
						stopQ = endPointX;
					if( (endPointY != 0) && (endPointY > stopQ) )
						stopQ = endPointY;
					if( (endPointZ != 0) && (endPointZ > stopQ) )
						stopQ = endPointZ;


					beginningPointX = endPointX = deleteFlagX = 0 ;
					beginningPointY = endPointY = deleteFlagY = 0 ;
					beginningPointZ = endPointZ = deleteFlagZ = 0 ;

				}


				doFlag = 0 ;
			}
			if( startQ < stopQ )
				jointPtr->interpolateQuaternion(startQ, stopQ) ;
		}
		//printf ("endif\n") ;
	}	

	//printf ("preinit\n") ;
	
	// Recompute the markers.
	mdata->recomputeEulerMarkers();
	mdata->recomputeQuaternionMarkers();

	// Re-init the graph editor.
	initGraphEditor (NULL, numNodeNames, 0) ;

	damage (1) ;
}


// This function is interactively called whilst dragging points around in the graph editor
// window.
void Gl_Graph_Window::dragPoints(int yDisplace)
{
	for (int i = 0; i < lineVector.size(); i += 2)
	{
		pVector tempPointVec = lineVector[i]   ;
		pVector tempSelMask  = lineVector[i+1] ;

		for (int j = 0; j < tempPointVec.size(); j++)
		{
			if (tempSelMask[j] == 1.0f)
			{
				tempPointVec[j] += yDisplace ;
			}
		}

		lineVector[i] = tempPointVec ;
	}
}

// This function is called after the draggig of the points is completed.  It re-initializes
// the proper joint objects and re-initilizes the graph editor.
void Gl_Graph_Window::doneDraggingPoints(void)
{
	char tempString[1000]   ;
	joint* jointPtr ;
	char * selection = bigNameString ;

	// Dig through what is and isn't selected and 'drag' their values accordingly.
	for (int k = 0; k < lineVector.size(); k += 2)
	{
		pVector tempPointVec = lineVector[k]   ;
		pVector tempSelMask  = lineVector[k+1] ;

		sscanf (selection, "%s", tempString) ;
		selection = selection + strlen(tempString) + 1 ;

		int doFlag = 0 ;

		if (isEffector[k] == 0)
		{
			for (int j = 0; j < tempPointVec.size(); j += ONE)
			{
				//If an X, Y, or Z value has been changed...
				if (tempSelMask[j] == 1.0f || tempSelMask[j+1] == 1.0f || tempSelMask[j+2] == 1.0f)
				{
					jointPtr = mdata->getJoint(tempString) ;
					jointPtr->setRotation((j/ONE + 1), tempPointVec[j], tempPointVec[j+1], tempPointVec[j+2]) ;
				}
			}
		}
	}	

	// Redo the Euler and quaternion markers after the manipulation is complete.
	mdata->recomputeEulerMarkers();
	mdata->recomputeQuaternionMarkers();

	//printf ("Done moving poitns around\n") ;

	initGraphEditor (NULL, numNodeNames, 1) ;
}
