/**************************************************************************************************
*
*	Files: mainWindow.h and mainWindow.cpp
*	Written by Steve Dutcher and Andy Gardner
*
*	This class handles the main window which has the buttons for what is displayed,
*	the time slider, menubar and animation window.  The user from this window can
*	also create a graphEditor window along with tearing off a copy of the animation
*	window, that will keep playing while open.
*
*	NOTE:  Read mainWindow.h a description of the mainWindow class
*
**************************************************************************************************/

#include <stdio.h>

#include "joint.h"
#include "graphEditor.h"
#include "mainWindow.h"

static RunSlider* rSlider;
static RunButton* rButton;

// This is export rib function it pops up a window that asks you for
// the frame range you want to write along with the resolution and a filename
// It will then write out a rib file for you
void exportRibButtonFn(motionData *mdata)
{
	Fl_Double_Window expRibWin(180, 50, 300, 300, "RIB export window") ;
		Fl_Button eggsitButton   (25, 10, 100, 25, "Exit Window")       ;
		Fl_Button cRibButton (150, 10, 100, 25, "Create RIB File")      ;
		Fl_Input  beginF     (85, 50, 50, 25, "Start Frame")       ; 
		Fl_Input  endF       (85, 90, 50, 25, "End Frame")         ; 
		Fl_Input  resX       (85, 130, 50, 25, "Resolution X")     ;
		Fl_Input  resY       (85, 170, 50, 25, "Resolution Y")     ;
		Fl_Input  fName      (85, 200, 200, 25, "Output File")         ;

	expRibWin.set_modal();
	expRibWin.show();
	expRibWin.end();

	int notdone = 1;

	//Init those fields before we loop...
	char tempBuff[15];
	tempBuff[0] = '\0'   ;								//If the intfield works it'd be great...
	sprintf( tempBuff, "%d", mdata->getNumberOfFrames() );
	endF.value(tempBuff) ;
	beginF.value("1")    ;

	resX.value("320") ;
	resY.value("240") ;

	fName.value("defaultRender")   ;

	// As long as the exportRIB button is shown, poll it to see what the user is clicking.
	while(expRibWin.shown() && notdone)
	{  
		Fl::wait();
		for(;;)
		{
			Fl_Widget *q = Fl::readqueue();
			if(!q)
				break ;
			else if( int(&eggsitButton) == int(q) ) // Exit the window
			{
				notdone = 0 ;
			}
			else if( int(&cRibButton) == int(q) )	// Export the file
			{
				notdone = 0 ;
				char *fileName = fl_file_chooser("Choose a RIB file to write","*.rib",NULL);
				if( fileName  && fileName[0]!='\0' )
					mdata->writeRIBfile(fileName, fName.value(), atoi(resX.value()), atoi(resY.value()), atoi(beginF.value()), atoi(endF.value())) ; //name, x, y, frameNumber
				else
					notdone = 1;
			}
		}
	}
}

// Close the mainWindow if the user presses exit
void exit_cb(Fl_Widget* myself, void*w){
	if (fl_ask("Are you sure you want to quit?")) 
		delete ((mainWindow*)w);
}

// This is just the callback which calls the export rib function
void export_rib_cb(Fl_Widget*, void *mdata){
	exportRibButtonFn((motionData*)mdata) ;
}

// This function gets two filenames the input file which specifies what to read in
// and also the filename to read out.  It then calls writeMarkersFile which is part
// of the motionData class
void export_marker_cb(Fl_Widget*, void *mdata){
	char *readFile;
	char *filename = fl_file_chooser("Choose a file with Marker names to open","*.*",NULL);
	if( filename == NULL )
		return;
	readFile = strdup (filename) ;

	filename = fl_file_chooser("Choose a marker file to create","*.mrk",NULL);
	if( filename == NULL  || filename[0]=='\0')
		return;
	((motionData*)mdata)->writeMarkersToFile(readFile, filename) ;
	delete []readFile;
}

// This is the callback for the menuitem to write a mel script it calls
// motionData writeMelScript(filename) to do its work
void export_mel_cb(Fl_Widget*, void *mdata){
	char *filename = fl_file_chooser("Choose a MEL file to create","*.mel",NULL);
	if( filename == NULL  || filename[0]=='\0')
		return;
	((motionData*)mdata)->writeMelScript(filename) ;
}

// This function dumps out the motionData to the screen so that you can analyze 
// individual values
void dump_motion_data_cb(Fl_Widget*, void* mdata){
	((motionData*)mdata)->dump();
}

// This function loads a bvh file by calling motionData openMotionDataFile(filename)
void load_bvh_cb(Fl_Widget *, void *aw){
	char *filename = fl_file_chooser("Choose a BVH file to open","*.bvh",NULL);
	if( filename == NULL || filename[0]=='\0' )
		return;

	((animWindow*)aw)->openMotionDataFile (filename) ;
}

// This is the callback for saving a bvh file 
void save_bvh_cb(Fl_Widget*, void *mdata){
	char *filename = fl_file_chooser("Save BVH file as","*.bvh",NULL);
	if( filename == NULL  || filename[0]=='\0')
		return;
	((motionData*)mdata)->writeBVHfile(filename) ;
}

// This is the call back for loading an acclaim file.  Two files need to
// be open an ASF file and an AMC file
void load_acclaim_cb(Fl_Widget* , void *aw){
	char *fileNameASF;
	char *filenameA = fl_file_chooser("Choose an ASF file to open","*.asf",NULL);
	if( filenameA == NULL )
		return;
	fileNameASF = strdup (filenameA) ;

	char *fileNameAMC = fl_file_chooser("Choose the matching AMC file","*.amc",NULL);
	if( fileNameAMC == NULL )
		return;

	if( fileNameASF && fileNameASF[0]!='\0' && fileNameAMC && fileNameAMC[0]!='\0' )
		((animWindow*)aw)->openAcclaimDataFiles (fileNameASF, fileNameAMC) ;

	delete []fileNameASF;
}

// This callback creates a new window that is the graphEditor
void graph_editor_cb(Fl_Widget*, void *mdata){
	new Graph_Editor_Window("Animation Graph Editor", (motionData*)mdata, rSlider);
}

// This callback creates a clone of the current animation window in the mainWindow
void animation_window_cb(Fl_Widget*, void *mw){
	((mainWindow*)mw)->cloneAnimationWindow();
}


/*********************************************************************************
* 
* This next group of callbacks are all for the buttons on the left side
* of mainWindow  they just call setShown so that the animWindow knows what
* things to shown and then it damages that window
*
*********************************************************************************/
void showOriginalMarker( Fl_Widget *myself, void *aw){
	((animWindow*)aw)->setShown( ((Fl_Button*)myself)->value(), ORIGINAL_MARKER);
	((animWindow*)aw)->damage(1);
}
void showOriginalLine( Fl_Widget *myself, void *aw){
	((animWindow*)aw)->setShown( ((Fl_Button*)myself)->value(), ORIGINAL_LINE);
	((animWindow*)aw)->damage(1);
}
void showOriginalEllipsoid( Fl_Widget *myself, void *aw){
	((animWindow*)aw)->setShown( ((Fl_Button*)myself)->value(), ORIGINAL_ELLIPSOID);
	((animWindow*)aw)->damage(1);
}
void showEulerMarker( Fl_Widget *myself, void *aw){
	((animWindow*)aw)->setShown( ((Fl_Button*)myself)->value(), EULER_MARKER);
	((animWindow*)aw)->damage(1);
}
void showEulerLine( Fl_Widget *myself, void *aw){
	((animWindow*)aw)->setShown( ((Fl_Button*)myself)->value(), EULER_LINE);
	((animWindow*)aw)->damage(1);
}
void showEulerEllipsoid( Fl_Widget *myself, void *aw){
	((animWindow*)aw)->setShown( ((Fl_Button*)myself)->value(), EULER_ELLIPSOID);
	((animWindow*)aw)->damage(1);
}
void showQuaternionMarker( Fl_Widget *myself, void *aw){
	((animWindow*)aw)->setShown( ((Fl_Button*)myself)->value(), QUATERNION_MARKER);
	((animWindow*)aw)->damage(1);
}
void showQuaternionLine( Fl_Widget *myself, void *aw){
	((animWindow*)aw)->setShown( ((Fl_Button*)myself)->value(), QUATERNION_LINE);
	((animWindow*)aw)->damage(1);
}
void showQuaternionEllipsoid( Fl_Widget *myself, void *aw){
	((animWindow*)aw)->setShown( ((Fl_Button*)myself)->value(), QUATERNION_ELLIPSOID);
	((animWindow*)aw)->damage(1);
}


// Here is the constructor for the mainWindow class
mainWindow::mainWindow(int x, int y, int w, int h, char* name, motionData *md) :
	Fl_Double_Window(x, y, w, h, name),
	menubar(0,0,w,30),

	text1(0,60,135,30, "Show Using"),
	text2(0,90,135,30, "Original Data"),
	omButton(0,100,135,30,"Effector Markers"),
	ojlButton(0,130,135,30,"Joint Lines" ),
	ojeButton(0,160,135,30, "Joint Ellipsoids"),
	
	text3(0,220,135,30,"Euler Data"),
	emButton(0,230,135,30,"Effector Markers"),
	ejlButton(0,260,135,30,"Joint Lines" ),
	ejeButton(0,290,135,30, "Joint Ellipsoids"),
	
	text4(0,350,135,30,"Quaternion Data"),
	qmButton(0,360,135,30,"Effector Markers"),
	qjlButton(0,390,135,30,"Joint Lines" ),
	qjeButton(0,420,135,30, "Joint Ellipsoids"),
	
	group(0,0,w,h)
{

	mdata = md;

	animWindow *aw = new animWindow(140, 30, w-140, h-60, mdata);

	if( rSlider == NULL ){
		// If slider has not yet been created then create it
		rs = new RunSlider(aw, mdata->getNumberOfFrames() ,60,h-30,w-60,30);
		aw->time = rs;
		rb = new RunButton(rs,0,h-30,60,30);
		aw->play = rb   ;
		rSlider = rs;
		rButton = rb;
		group.add(rButton);
		group.add(rSlider);
		group.add(text1);
		group.add(text2);
		group.add(text3);
		group.add(text4);
	}else{
		aw->time = rSlider;
		aw->play = rButton;
		aw->resize(0,30,w,h-30);
	}
	
	group.add(aw);

	resizable(group);

	// Set up menubar items
	menubar.add("File/Open &BVH File",FL_ALT+'b',load_bvh_cb,(void*)aw);
	menubar.add("File/Open *Acclaim File(s)",FL_ALT+'a',load_acclaim_cb,(void*)aw,FL_MENU_DIVIDER);
	menubar.add("File/Save &BVH File",FL_ALT+'s',save_bvh_cb,(void*)mdata,FL_MENU_DIVIDER);
	menubar.add("File/Export &Marker Data",FL_ALT+'m',export_marker_cb,(void*)mdata);
	menubar.add("File/Export &RIB File",FL_ALT+'r',export_rib_cb,(void*)mdata);
	menubar.add("File/Export &MEL File",FL_ALT+'l',export_mel_cb,(void*)mdata, FL_MENU_DIVIDER);
	menubar.add("File/Dump Motion Data",0,dump_motion_data_cb,(void*)mdata,FL_MENU_DIVIDER);
	menubar.add("File/&Exit",FL_ALT+'e',exit_cb,(void*)this);
	menubar.add("Window/Clone Animation Window",0,animation_window_cb,(void*)this);
	menubar.add("Window/Graph Editor",0,graph_editor_cb,(void*)mdata);

	// Set color and callback for the ORIGINAL shown buttons
	omButton.selection_color( FL_RED );
	omButton.callback( (Fl_Callback*)showOriginalMarker, (void*)aw );
	ojlButton.selection_color( FL_BLUE );
	ojlButton.callback( (Fl_Callback*)showOriginalLine, (void*)aw );
	ojeButton.selection_color( FL_CYAN );
	ojeButton.callback( (Fl_Callback*)showOriginalEllipsoid, (void*)aw );

	// Set color and callback for the EULER shown buttons
	emButton.selection_color( FL_GREEN );
	emButton.callback( (Fl_Callback*)showEulerMarker, (void*)aw );
	ejlButton.selection_color( FL_RED );
	ejlButton.callback( (Fl_Callback*)showEulerLine, (void*)aw );
	ejeButton.selection_color( FL_MAGENTA );
	ejeButton.callback( (Fl_Callback*)showEulerEllipsoid, (void*)aw );

	// Set color and callback for the QUATERNION shown buttons
	qmButton.selection_color( FL_BLUE );
	qmButton.callback( (Fl_Callback*)showQuaternionMarker, (void*)aw );
	qjlButton.selection_color( FL_GREEN );
	qjlButton.callback( (Fl_Callback*)showQuaternionLine, (void*)aw );
	qjeButton.selection_color( FL_YELLOW );
	qjeButton.callback( (Fl_Callback*)showQuaternionEllipsoid, (void*)aw );

	emButton.value(1);
	ejlButton.value(1);
	
	end();
	show();
}

void mainWindow::cloneAnimationWindow()
{
	// Copy all the values of the current animWindow into the array shownValues
	// and then pass that function to the new animation window that is created so
	// that it knows what to show
	int shownValues[9];
	shownValues[0] = omButton.value();
	shownValues[1] = ojlButton.value();
	shownValues[2] = ojeButton.value();
	shownValues[3] = emButton.value();
	shownValues[4] = ejlButton.value();
	shownValues[5] = ejeButton.value();
	shownValues[6] = qmButton.value();
	shownValues[7] = qjlButton.value();
	shownValues[8] = qjeButton.value();
	
	Fl_Double_Window *w = new Fl_Double_Window(0,0,500,500,"Secondary Animation Window" );
	Fl_Group *g = new Fl_Group(0,0,500,500);
	w->begin();
	animWindow *a = new animWindow(0,0,500,500,mdata, shownValues);

	// Give animWindow a pointer to the slider so that it can get the time from it
	a->time = rSlider;
	a->play = rButton;
	w->resizable(g);
	w->end();
	w->show();

}

mainWindow::~mainWindow()
{
}
